import typing

from ctc import spec
import numpy as np
import pandas as pd


def create_timestamp_column(
    df: spec.DataFrame,
    timestamps: typing.Sequence[int],
) -> spec.Series:
    """create a timestamp column series that shares an index with dataframe"""

    index = df.index

    if isinstance(index, pd.MultiIndex):
        # keep only first level of multiindex
        n_levels = len(df.index.names)
        df = typing.cast(spec.DataFrame, df.droplevel(list(range(1, n_levels))))

    merged = pd.merge(
        df,
        pd.Series(timestamps, name='timestamp'),
        left_index=True,
        right_index=True,
        how='left',
    )
    merged.index = index

    return merged['timestamp']


def create_datetime_column(df: spec.DataFrame, timestamps=None) -> spec.Series:
    if 'timestamp' in df.columns:
        timestamp_column = df['timestamp']
    else:
        if timestamps is None:
            raise Exception('must specify timestamps')
        timestamp_column = create_timestamp_column(df, timestamps)
    return pd.to_datetime(timestamp_column, unit='s')


def create_date_column(df, timestamps=None):
    if 'datetime' in df.columns:
        datetime = df['datetime']
    else:
        if timestamps is None:
            raise Exception('must specify timestamps')
        datetime = create_timestamp_column(df, timestamps)
    return datetime.dt.date


def create_week_column(df: spec.DataFrame, timestamps: typing.Sequence[int]):
    if 'datetime' in df.columns:
        datetime = df['datetime']
    else:
        datetime = create_timestamp_column(df, timestamps)
    year = datetime.dt.isocalendar().year.astype(str)
    week = datetime.dt.isocalendar().week.astype(str)
    return year + '-' + week


def add_missing_series_dates(
        series: spec.Series, datetimes: spec.Series, fill_value: int = 0
) -> spec.Series:
    series = series.copy()
    all_days = np.arange(
        datetimes.min().timestamp(),
        datetimes.max().timestamp(),
        86400,
    )
    for day in all_days:
        date = pd.to_datetime(day, unit='s').date()
        if date not in series:
            series[date] = fill_value
    series = series.sort_index()
    return series


def add_missing_series_weeks(
    series: spec.Series, datetimes: spec.Series, fill_value: int = 0
) -> spec.Series:

    series = series.copy()

    first_year = datetimes.dt.isocalendar().year.min()
    last_year = datetimes.dt.isocalendar().year.max()

    for year in range(first_year, last_year + 1):

        # get first week of year
        if year == first_year:
            first_week = datetimes.dt.isocalendar().week.min()
        else:
            first_week = 1

        # get last week of year
        if year == last_year:
            last_week = datetimes.dt.isocalendar().week.max()
        else:
            last_week = 52

        # iterate
        for week in list(range(first_week, last_week + 1)):
            week_str = str(year) + '-' + str(week)
            if week_str not in series:
                series[week_str] = fill_value

    series = series.sort_index()

    return series




from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_eth_gas_price(provider=None, decode_response=True):
    request = rpc_constructors.construct_eth_gas_price()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_gas_price(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_accounts(provider=None):
    request = rpc_constructors.construct_eth_accounts()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_accounts(response=response)


async def async_eth_sign(address, message, provider=None):
    request = rpc_constructors.construct_eth_sign(
        address=address,
        message=message,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_sign(response=response)


async def async_eth_sign_transaction(
    from_address,
    data,
    to_address=None,
    gas=None,
    gas_price=None,
    value=None,
    nonce=None,
    provider=None,
    snake_case_response=True,
):
    request = rpc_constructors.construct_eth_sign_transaction(
        from_address=from_address,
        data=data,
        to_address=to_address,
        gas=gas,
        gas_price=gas_price,
        value=value,
        nonce=nonce,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_sign_transaction(
        response=response,
        snake_case_response=snake_case_response,
    )


async def async_eth_send_transaction(
    from_address,
    data,
    to_address=None,
    gas=None,
    gas_price=None,
    value=None,
    nonce=None,
    provider=None,
    snake_case_response=True,
):
    request = rpc_constructors.construct_eth_send_transaction(
        from_address=from_address,
        data=data,
        to_address=to_address,
        gas=gas,
        gas_price=gas_price,
        value=value,
        nonce=nonce,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_send_transaction(
        response=response,
        snake_case_response=snake_case_response,
    )


async def async_eth_send_raw_transaction(data, provider=None):
    request = rpc_constructors.construct_eth_send_raw_transaction()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_send_raw_transaction(response=response)




import pytest

from ctc import evm


token_address = '0x956F47F50A910163D8BF957Cf5846D573E7f87CA'


@pytest.mark.asyncio
async def test_fetch_token_total_supply():
    total_supply = await evm.async_get_erc20_total_supply(
        token=token_address, block=13436621, normalize=False
    )
    assert total_supply == 542154246176449338629996405


@pytest.mark.asyncio
async def test_fetch_token_balance_of():
    address = '0x94b0a3d511b6ecdb17ebf877278ab030acb0a878'
    balance = await evm.async_get_erc20_balance_of(
        address=address, token=token_address, block=13436621, normalize=False
    )
    assert balance == 137680097380114452997476874




from __future__ import annotations

import typing

from ctc import spec

from .. import hashes


def get_function_parameter_types(
    function_abi: spec.FunctionABI = None,
    function_signature: typing.Optional[str] = None,
) -> list[spec.ABIDatumType]:

    if function_abi is not None:
        import eth_utils

        output = []
        for item in function_abi.get('inputs', []):
            cast_item = typing.cast(typing.Dict[str, typing.Any], item)
            collapsed = eth_utils.abi.collapse_if_tuple(cast_item)
            output.append(collapsed)

        return output

    elif function_signature is not None:
        parameter_str = function_signature.split('(')[1]
        parameter_str = parameter_str[:-1]
        return parameter_str.split(',')

    else:
        raise Exception('must specify function_abi or function_signature')


@typing.overload
def get_function_parameter_names(
    function_abi: spec.FunctionABI,
    require_names: typing.Literal[True],
) -> list[str]:
    ...


@typing.overload
def get_function_parameter_names(
    function_abi: spec.FunctionABI,
    require_names: typing.Literal[False] = False,
) -> list[typing.Optional[str]]:
    ...


def get_function_parameter_names(
    function_abi: spec.FunctionABI,
    require_names: bool = False,
) -> typing.Union[list[str], list[typing.Optional[str]]]:

    names = [item.get('name') for item in function_abi.get('inputs', [])]

    if require_names:
        if any(name is None for name in names):
            raise Exception('function abi does not specify names')
        else:
            return typing.cast(typing.List[str], names)
    else:
        return names


def get_function_signature(
    function_abi: spec.FunctionABI = None,
    parameter_types: typing.Optional[list[str]] = None,
    function_name: typing.Optional[str] = None,
    include_names: bool = False,
) -> str:

    # get parameter types
    if parameter_types is None:
        parameter_types = get_function_parameter_types(function_abi)
    parameter_types = [
        get_function_selector_type(item) for item in parameter_types
    ]

    # create entires with or without names
    if include_names:
        if function_abi is None:
            raise Exception('must specify function_abi')
        parameter_names = get_function_parameter_names(
            function_abi, require_names=True
        )
        entries = [
            type + ' ' + name
            for type, name in zip(parameter_types, parameter_names)
        ]
        entries_str = ', '.join(entries)
    else:
        entries = parameter_types
        entries_str = ','.join(entries)

    if function_name is None:
        if function_abi is None:
            raise Exception('must specify function_abi')
        function_name = function_abi['name']

    return function_name + '(' + entries_str + ')'


def get_function_selector_type(
    datatype: spec.ABIDatumType,
) -> spec.ABIDatumType:
    """get the data name used for function selector"""
    if datatype == 'uint':
        return 'uint256'
    elif datatype == 'int':
        return 'int256'
    elif datatype == 'ufixed':
        return 'ufixed128x18'
    elif datatype == 'fixed':
        return 'fixed128x18'
    else:
        return datatype


def get_function_selector(
    function_abi: typing.Optional[spec.FunctionABI] = None,
    function_signature: typing.Optional[spec.FunctionSignature] = None,
) -> str:

    if function_signature is None:
        function_signature = get_function_signature(function_abi)

    full_hash = hashes.keccak(function_signature.encode(), output_format='raw_hex')
    return full_hash[:8]


def get_function_output_types(
    function_abi: spec.FunctionABI,
) -> list[spec.ABIDatumType]:

    import eth_utils

    output_types = []
    for output in function_abi['outputs']:
        cast_output = typing.cast(typing.Dict[str, typing.Any], output)
        output_type = eth_utils.abi.collapse_if_tuple(cast_output)
        output_types.append(output_type)
    return output_types


def get_function_output_names(
    function_abi: spec.FunctionABI,
    human_readable: bool = False,
) -> list[typing.Optional[str]]:

    output_names = [output.get('name') for output in function_abi['outputs']]

    # human readable uses function name or output_# if no names provided
    if human_readable:
        if len(output_names) == 1:
            if output_names[0] == '':
                output_names[0] = function_abi.get('name')
        else:
            for on, output_name in list(enumerate(output_names)):
                if output_name == '':
                    output_names[on] == 'output_' + str(on)

    return output_names




from ctc import spec


def digest_eth_get_compilers(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_compile_lll(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_compile_solidity(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_compile_serpent(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response




import setuptools


setuptools.setup(
    name='checkthechain',
    version='0.2.3',
    packages=setuptools.find_packages("./src"),
    package_dir={'': 'src'},
    package_data={'ctc': ['default_data/*']},
    install_requires=[
        #
        # data science
        'matplotlib',
        'numpy',
        'pandas',
        'pandas-stubs',
        #
        # data dependencies
        'aiohttp',
        'aiofiles',
        'pyyaml',
        'toml',
        #
        # tool suite
        'toolcache',
        'toolcli',
        'toolconf',
        'toolplot',
        'toolstr',
        'tooltable',
        'tooltime',
        #
        # EVM dependencies
        'pycryptodome',  # for keccak()
        'eth_abi',  # for encode_single()/decode_single()
        'eth_utils',  # for collapse_if_tuple()
        'rlp',  # for create2 address computation
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-asyncio',
        ],
        'speed': [
            'pysha3',  # for keccak()
        ],
    },
    scripts=[
        './scripts/ctc',
    ],
)




from ctc import spec
from .. import rpc_request


def construct_eth_get_work() -> spec.RpcRequest:
    return rpc_request.create('eth_getWork', [])


def construct_eth_submit_work(
    nonce: spec.BinaryData,
    pow_hash: spec.BinaryData,
    digest: spec.BinaryData,
) -> spec.RpcRequest:
    parameters = [nonce, pow_hash, digest]
    return rpc_request.create('eth_submitWork', parameters)


def construct_eth_submit_hashrate(
    hashrate: spec.BinaryData, id: str
) -> spec.RpcRequest:
    return rpc_request.create('eth_submitHashrate', [hashrate, id])


def construct_eth_coinbase() -> spec.RpcRequest:
    return rpc_request.create('eth_coinbase', [])


def construct_eth_mining() -> spec.RpcRequest:
    return rpc_request.create('eth_mining', [])


def construct_eth_hashrate() -> spec.RpcRequest:
    return rpc_request.create('eth_hashrate', [])




from __future__ import annotations

from ctc import spec
from .. import rpc_request


def construct_shh_version() -> spec.RpcRequest:
    return rpc_request.create('shh_version', [])


def construct_shh_post(
    from_address: spec.BinaryData,
    to_address: spec.BinaryData,
    topics: list[spec.BinaryData],
    payload: spec.BinaryData,
    priority: spec.BinaryData,
    ttl: spec.BinaryData,
) -> spec.RpcRequest:
    data: dict = {
        'from': from_address,
        'to': to_address,
        'topics': topics,
        'payload': payload,
        'priority': priority,
        'ttl': ttl,
    }
    data = {k: v for k, v in data.items() if v is not None}
    return rpc_request.create('shh_new_filter', [data])


def construct_shh_new_identity() -> spec.RpcRequest:
    return rpc_request.create('ssh_new_version', [])


def construct_shh_has_identity(data: spec.BinaryData) -> spec.RpcRequest:
    return rpc_request.create('shh_has_identity', [data])


def construct_shh_new_group() -> spec.RpcRequest:
    return rpc_request.create('shh_new_group', [])


def construct_shh_add_to_group(data: spec.BinaryData) -> spec.RpcRequest:
    return rpc_request.create('shh_add_to_group', [data])


def construct_shh_new_filter(
    to_address: spec.BinaryData, topics: list[spec.BinaryData]
) -> spec.RpcRequest:
    payload = {
        'to': to_address,
        'topics': topics,
    }
    return rpc_request.create('shh_new_filter', [payload])


def construct_shh_uninstall_filter(
    filter_id: spec.BinaryData,
) -> spec.RpcRequest:
    return rpc_request.create('shh_uninstall_filter', [filter_id])


def construct_shh_get_filter_changes(
    filter_id: spec.BinaryData,
) -> spec.RpcRequest:
    return rpc_request.create('shh_get_filter_changes', [filter_id])


def construct_shh_get_messages(filter_id: spec.BinaryData) -> spec.RpcRequest:
    return rpc_request.create('shh_get_messages', [filter_id])




import pytest

from ctc import rpc


@pytest.mark.asyncio
async def test_eth_gas_price():
    result = await rpc.async_eth_gas_price()
    assert isinstance(result, int)


@pytest.mark.asyncio
async def test_eth_accounts():
    result = await rpc.async_eth_accounts()
    assert isinstance(result, list)




from __future__ import annotations

import typing

from ctc import spec

from .. import formats
from . import function_parsing
from . import event_parsing


def get_function_abi(
    contract_abi: spec.ContractABI,
    function_name: typing.Optional[str] = None,
    n_parameters: typing.Optional[int] = None,
    parameter_types: typing.Optional[list[spec.ABIDatumType]] = None,
    function_selector: typing.Optional[spec.FunctionSelector] = None,
) -> spec.FunctionABI:

    if function_selector is not None:
        function_selector = formats.convert(function_selector, 'prefix_hex')

    candidates = []
    for item in contract_abi:
        if item.get('type') != 'function':
            continue
        else:
            function_abi = typing.cast(spec.FunctionABI, item)

        if (
            function_name is not None
            and function_abi.get('name') != function_name
        ):
            continue
        if (
            n_parameters is not None
            and len(function_abi['inputs']) != n_parameters
        ):
            continue
        if parameter_types is not None:
            types = function_parsing.get_function_parameter_types(function_abi)
            if parameter_types != tuple(types):
                continue
        if function_selector is not None:
            item_selector = function_parsing.get_function_selector(function_abi)
            item_selector = formats.convert(item_selector, 'prefix_hex')
            if item_selector != function_selector:
                continue
        candidates.append(function_abi)

    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) == 0:
        raise LookupError('could not find function abi')
    elif len(candidates) > 0:
        import json

        as_json = [json.dumps(candidate) for candidate in candidates]
        if all(as_json[0] == entry for entry in as_json[1:]):
            return candidates[0]

        raise LookupError('too many candidates found for function abi')
    else:
        raise Exception('internal error')


def get_event_abi(
    *,
    contract_abi: spec.ContractABI,
    event_name: typing.Optional[str] = None,
    event_hash: typing.Optional[str] = None,
    event_abi: typing.Optional[spec.EventABI] = None,
) -> spec.EventABI:
    """get event abi from contract abi"""

    if event_abi is not None:
        return event_abi

    if event_name is None and event_hash is None:
        raise Exception('specify event_name or event_hash')

    candidates = []
    for item in contract_abi:
        if item['type'] != 'event':
            continue
        if event_name is not None and item.get('name') != event_name:
            continue
        if event_hash is not None:
            item_hash = event_parsing.get_event_hash(event_abi=item)
            if item_hash != event_hash:
                continue

        candidates.append(item)

    if len(candidates) == 0:
        raise Exception('could not find event abi')
    elif len(candidates) == 1:
        return candidates[0]
    else:
        raise Exception('found too many candidates for event abi')




import asyncio

from ctc import rpc
from ctc.protocols import rari_utils


def get_command_spec():
    return {
        'f': fuse_command,
        'args': [
            {'name': 'pool', 'kwargs': {'nargs': '?'}},
            {'name': '--block', 'kwargs': {'default': 'latest'}},
        ],
    }


def fuse_command(pool, block, **kwargs):
    if block != 'latest':
        block = int(block)

    if pool is not None:

        try:
            pool_index = int(pool)
            arg_type = 'pool_index'
        except Exception as e:
            token = pool
            arg_type = 'token'

        if arg_type == 'pool_index':
            asyncio.run(summarize_fuse_pool(pool_index=pool_index, block=block))
        elif arg_type == 'token':
            asyncio.run(summarize_fuse_token(token, block))

    else:
        asyncio.run(summarize_all_fuse_pools(block=block))


async def summarize_fuse_pool(pool_index, block):
    all_pools = await rari_utils.async_get_all_pools(block=block)
    pool = all_pools[pool_index]
    comptroller = pool[2]
    block_data = await rpc.async_eth_get_block_by_number(block_number=block)

    tokens_data = await rari_utils.async_get_pool_summary(
        comptroller=comptroller,
        block=block,
    )
    pool_name = await rari_utils.async_get_pool_name(comptroller=comptroller)
    rari_utils.print_fuse_pool_summary(
        block=block_data,
        tokens_data=tokens_data,
        pool_name=pool_name,
    )

    from ctc.rpc.rpc_backends import rpc_http_async

    provider = rpc.get_provider()
    await rpc_http_async.async_close_http_session(provider=provider)


async def summarize_all_fuse_pools(block):

    await rari_utils.print_all_pool_summary(block=block)

    from ctc.rpc.rpc_backends import rpc_http_async

    provider = rpc.get_provider()
    await rpc_http_async.async_close_http_session(provider=provider)


async def summarize_fuse_token(token, block):
    if token == 'ETH':
        token = '0x0000000000000000000000000000000000000000'
    await rari_utils.async_print_fuse_token_summary(token=token, block=block)

    from ctc.rpc.rpc_backends import rpc_http_async

    provider = rpc.get_provider()
    await rpc_http_async.async_close_http_session(provider=provider)




from __future__ import annotations

import asyncio
import logging
import random

import aiohttp

from ctc import spec
from .. import rpc_provider


_http_sessions: dict[spec.ProviderKey, aiohttp.ClientSession] = {}


async def async_send_http(
    request: spec.RpcRequest,
    provider: spec.ProviderSpec,
    n_attempts: int = 8,
) -> spec.RpcResponse:
    provider = rpc_provider.get_provider(provider)
    session = get_async_http_session(provider=provider)

    for attempt in range(n_attempts):

        async with session.post(provider['url'], json=request) as response:
            if response.status != 200:
                t_sleep = 2 ** attempt + random.random()
                print('sleeping for ' + str(t_sleep))
                await asyncio.sleep(t_sleep)
                continue
            return await response.json()

    else:
        message = (
            'http rpc request failed after '
            + str(n_attempts)
            + ' retries, status_code = '
            + str(response.status)
        )
        logger = logging.getLogger()
        logger.info(message)
        raise Exception(message)


def get_async_http_session(
    provider: spec.Provider, create: bool = True
) -> aiohttp.ClientSession:
    key = rpc_provider.get_provider_key(provider)
    if key not in _http_sessions:
        if create:
            kwargs = provider['session_kwargs']
            if kwargs is None:
                kwargs = {}
            _http_sessions[key] = aiohttp.ClientSession(**kwargs)
        else:
            raise Exception('no session, must create')
    return _http_sessions[key]


async def async_close_http_session(provider: spec.ProviderSpec = None) -> None:
    provider = rpc_provider.get_provider(provider)
    session = get_async_http_session(provider=provider)
    await asyncio.sleep(0)
    await session.close()




from __future__ import annotations

import typing

from ctc import spec
from .. import formats
from . import abi_coding
from . import event_parsing


def decode_event_topics(
    topics: typing.Sequence[spec.BinaryData],
    event_abi: typing.Optional[spec.EventABI] = None,
    indexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    indexed_names: typing.Optional[list[str]] = None,
    use_names: bool = True,
):
    """
    remaining edgecase:
    - variable data in indexed topic?
    """

    # get abi
    if indexed_types is None:
        if event_abi is None:
            raise Exception('must specify event_abi')
        indexed_types = event_parsing.get_event_indexed_types(event_abi)

    # decode
    decoded_topics = []
    for topic, indexed_type in zip(topics[1:], indexed_types):
        topic = formats.convert(topic, 'binary')
        decoded_topic = abi_coding.abi_decode(topic, indexed_type)
        decoded_topics.append(decoded_topic)

    # package output
    if not use_names:
        return decoded_topics
    else:
        if indexed_names is None:
            if event_abi is None:
                raise Exception('must specify event_abi')
            indexed_names = event_parsing.get_event_indexed_names(event_abi)
        return dict(zip(indexed_names, decoded_topics))


def decode_event_unindexed_data(
    data: spec.BinaryData,
    event_abi: typing.Optional[spec.EventABI] = None,
    unindexed_types: typing.Optional[list[spec.ABIDatumType]] = None,
    unindexed_names: typing.Optional[list[str]] = None,
    use_names: bool = True,
):
    """decode the unindexed data of event"""

    # gather metadata
    if unindexed_types is None:
        if event_abi is None:
            raise Exception('must specify event_abi')
        unindexed_types = event_parsing.get_event_unindexed_types(event_abi)

    # decode data
    data = formats.convert(data, 'binary')
    decoded = abi_coding.abi_decode(data, '(' + ','.join(unindexed_types) + ')')

    # package outputs
    if not use_names:
        return decoded
    else:
        if unindexed_names is None:
            if event_abi is None:
                raise Exception('must specify event_abi')
            unindexed_names = event_parsing.get_event_unindexed_names(event_abi)
        return dict(zip(unindexed_names, decoded))


def normalize_event(
    event: spec.RawLog,
    event_abi: spec.EventABI,
    arg_prefix: str = 'arg__',
) -> spec.NormalizedLog:

    # decode event args
    decoded_topics = decode_event_topics(
        topics=event['topics'], event_abi=event_abi
    )
    decoded_data = decode_event_unindexed_data(
        data=event['data'], event_abi=event_abi
    )

    # remove keys
    remove_keys = ['data', 'topics', 'removed']
    normalized = {k: v for k, v in event.items() if k not in remove_keys}

    # rename keys
    normalized['contract_address'] = normalized['address']

    # add additional keys
    normalized['event_name'] = event_abi['name']
    normalized['event_hash'] = event['topics'][0]

    # add event args
    if arg_prefix is None:
        arg_container = {}
        normalized['args'] = arg_container
        arg_prefix = ''
    else:
        arg_container = normalized
    for event_args in [decoded_topics, decoded_data]:
        for arg_name, arg_value in event_args.items():
            key = arg_prefix + arg_name
            if key in arg_container:
                raise Exception('event key collision: ' + str(key))
            arg_container[key] = arg_value

    return normalized


#
# # dataframes
#


def decode_events_dataframe(
    df: spec.DataFrame,
    event_abi: spec.EventABI,
    delete_data_column: bool = True,
) -> spec.DataFrame:
    """decode dataframe that contains raw event data

    ## Replaces Columns
    - topic0
    - topic1
    - topic2
    - topic3

    ## Removes Columns
    - data

    ## Adds Columns
    - arg__<name>
    """

    raise Exception('need to figure out if this could be rewritten better')

    # validate data
    if len(df) > 0:

        # assert only one event type used
        assert len(set(df['topic0'])) == 1

        # assert event hash matches event_abi
        assert df['topic0'].iloc[0] == event_parsing.get_event_hash(
            event_abi=event_abi
        )

    # decode data items
    unindexed_types = event_parsing.get_event_unindexed_types(
        event_abi=event_abi
    )
    unindexed_names = event_parsing.get_event_unindexed_names(
        event_abi=event_abi
    )
    for name in unindexed_names:
        if name in df.columns:
            raise Exception('column name collision')
    decoded = {name: {} for name in unindexed_names}
    for index, value in df['data'].items():
        decoded_data = decode_event_unindexed_data(
            data=value,
            unindexed_types=unindexed_types,
            unindexed_names=unindexed_names,
        )
        for name, subvalue in decoded_data.items():
            decoded[name][index] = subvalue

    # concatenate
    import pandas as pd

    new_df = pd.concat([df, pd.DataFrame(decoded)], axis=1)
    new_df = new_df.reindex(df.index)
    if delete_data_column:
        del new_df['data']

    # rename topic0 to event_type
    if 'event_type' in new_df:
        raise Exception('column name collision')
    new_df = new_df.rename(columns={'topic0': 'event_type'})

    # decode other topics
    # need to double check this section
    indexed_types = event_parsing.get_event_indexed_types(event_abi=event_abi)
    for t, indexed_type in enumerate(indexed_types):
        if indexed_type == 'address':
            topic = 'topic' + str(t + 1)
            new_df[topic] = '0x' + new_df[topic].str[26:]

    # rename topicX to indexed variable names
    indexed_names = event_parsing.get_event_indexed_names(event_abi=event_abi)
    indexed_rename = {
        'topic' + str(i + 1): name for i, name in enumerate(indexed_names)
    }
    for name in indexed_names:
        if name in new_df.columns:
            raise Exception('column name collision')
    new_df = new_df.rename(columns=indexed_rename)
    if len(indexed_names) == 1:
        del new_df['topic2']
        del new_df['topic3']
    elif len(indexed_names) == 2:
        del new_df['topic3']

    return new_df




from ctc import binary
from ctc import spec
from .. import rpc_format
from .. import rpc_request


def construct_eth_block_number() -> spec.RpcRequest:
    return rpc_request.create('eth_blockNumber', [])


def construct_eth_get_block_by_hash(
    block_hash: spec.BinaryData, include_full_transactions: bool = False
) -> spec.RpcRequest:
    encoded_block_hash = binary.convert(block_hash, 'prefix_hex')
    parameters = [encoded_block_hash, include_full_transactions]
    return rpc_request.create('eth_getBlockByHash', parameters)


def construct_eth_get_block_by_number(
    block_number: spec.BlockSpec,
    include_full_transactions: bool = True,
) -> spec.RpcRequest:

    encoded_block_number = rpc_format.encode_block_number(block_number)

    parameters = [encoded_block_number, include_full_transactions]
    return rpc_request.create(
        method='eth_getBlockByNumber',
        parameters=parameters,
    )


def construct_eth_get_uncle_count_by_block_hash(
    block_hash: spec.BinaryData,
) -> spec.RpcRequest:
    encoded_block_hash = binary.convert(block_hash, 'prefix_hex')
    return rpc_request.create(
        method='eth_getUncleCountByBlockHash',
        parameters=[encoded_block_hash],
    )


def construct_eth_get_uncle_count_by_block_number(
    block_number: spec.BlockSpec,
) -> spec.RpcRequest:
    encoded_block_number = rpc_format.encode_block_number(block_number)
    return rpc_request.create(
        method='eth_getUncleCountByBlockNumber',
        parameters=[encoded_block_number],
    )


def construct_eth_get_uncle_by_block_hash_and_index(
    block_hash: spec.BinaryData, uncle_index: spec.BinaryData
) -> spec.RpcRequest:

    encoded_block_hash = binary.convert(block_hash, 'prefix_hex')
    encoded_uncle_index = binary.convert(uncle_index, 'prefix_hex')

    return rpc_request.create(
        method='eth_getUncleByBlockHashAndIndex',
        parameters=[encoded_block_hash, encoded_uncle_index],
    )


def construct_eth_get_uncle_by_block_number_and_index(
    block_number: spec.BlockSpec, uncle_index: spec.BinaryData
) -> spec.RpcRequest:

    encoded_block_number = rpc_format.encode_block_number(block_number)
    encoded_uncle_index = binary.convert(uncle_index, 'prefix_hex')

    return rpc_request.create(
        method='eth_getUncleByBlockNumberAndIndex',
        parameters=[encoded_block_number, encoded_uncle_index],
    )





def calc_out_given_in(
    token_amount_in,
    token_balance_in,
    token_weight_in,
    token_balance_out,
    token_weight_out,
    swap_fee,
):
    weight_ratio = token_weight_in / token_weight_out
    adjusted_in = token_amount_in * (1 - swap_fee)
    y = token_balance_in / (token_balance_in + adjusted_in)
    foo = pow(y, weight_ratio)
    bar = 1 - foo
    token_amount_out = token_balance_out * bar
    return token_amount_out




import pytest

from ctc import evm


fei_address = '0x956F47F50A910163D8BF957Cf5846D573E7f87CA'


def test_get_token_address():
    assert fei_address.lower() == evm.get_erc20_address('FEI').lower()


@pytest.mark.asyncio
async def test_fetch_token_decimals():
    assert 18 == await evm.async_get_erc20_decimals(fei_address)


@pytest.mark.asyncio
async def test_fetch_token_name():
    assert 'Fei USD' == await evm.async_get_erc20_name(fei_address)


@pytest.mark.asyncio
async def test_fetch_token_symbol():
    assert 'FEI' == await evm.async_get_erc20_symbol(fei_address)




from ctc import spec

# import websockets

# from .. import rpc_provider


async def async_send_websocket(
    request: spec.RpcRequest,
    provider: spec.Provider,
) -> spec.RpcResponse:
    raise NotImplementedError()
    # provider = rpc_provider.get_provider(provider)
    # async with websockets.connect(provider['url']) as websocket:
    #     await websocket.send(request)
    #     return await websocket.recv()




# from ctc.cli import cli_run


def get_command_spec():
    return {
        'f': root_command,
    }


available_subcommands = """
    cd                  change working directory to a ctc directory
    config              print current ctc config
    config edit         edit current ctc config
    config path         print current ctc config path
    download-proxy-abi  download a proxy implementation of an abi
    setup               run ctc setup wizard
    rechunk-events      rechunk how events are stored on disk
    keccak              compute keccak hash of a hex or text string
    address             print summary of an address, includin abi
    block               print summmary of a block
    blocks              export block data
    call                print output of a contract call
    calls               export calls across contracts or blocks
    eth balance         print eth balance of a single wallet
    eth balances        export eth balances across wallets or blocks
    erc20 balance       print an erc20 balance of a single wallet
    erc20 balances      export balances of erc20s, blocks, or addresses
    erc20 transfers     export transfers of an erc20
    events              export events
    gas                 export gas pricing statistics
    transaction         print summary of a transaction
    chainlink           export chainlink feed data
    fei payload         export fei collateralization data
    rari fuse           export rari
    uniswap mints       export mints of a uniswap pool
    uniswap burns       export burns of a uniswap pool
    uniswap swaps       export swaps of a uniswap pool
    uniswap pool        print summary of a uniswap pool"""


def root_command():
    print('check the chain')
    print()
    print('usage:')
    print('    ctc <subcommands> [options]')
    print()
    print('available subcommands:')
    # for command in cli_run.command_index.keys():
    #     if command == ():
    #         continue
    #     print('   ', ' '.join(command))
    print()
    print(available_subcommands)




"""

information:
- pool metadata
- pool state
"""
import toolstr

from ctc.protocols import uniswap_v2_utils
from ctc.toolbox.amm_utils import cpmm
from ctc import rpc


def get_command_spec():
    return {
        'f': async_pool_command,
        'args': [
            {'name': 'pool'},
            {'name': '--block', 'kwargs': {'default': 'latest'}},
        ],
    }


async def async_pool_command(pool, block):
    tokens_metadata = await uniswap_v2_utils.async_get_pool_tokens_metadata(pool)
    x_symbol = tokens_metadata['x_symbol']
    y_symbol = tokens_metadata['y_symbol']
    title = 'Uniswap V2 Pool: ' + x_symbol + ' x ' + y_symbol
    toolstr.print_text_box(title)
    print('-', x_symbol, 'address:', tokens_metadata['x_address'])
    print('-', y_symbol, 'address:', tokens_metadata['y_address'])
    print('-', x_symbol, 'decimals:', tokens_metadata['x_decimals'])
    print('-', y_symbol, 'decimals:', tokens_metadata['y_decimals'])
    print()
    print()

    toolstr.print_header('Pool State')
    pool_state = await uniswap_v2_utils.async_get_pool_state(pool, block=block)
    cpmm.print_pool_summary(
        x_name=x_symbol,
        y_name=y_symbol,
        **pool_state,
    )

    await rpc.async_close_http_session()




from ctc import rpc


async def async_is_contract_address(address, block='latest', provider=None):
    code = await rpc.async_eth_get_code(
        address=address, block_number=block, provider=provider,
    )
    return len(code) > 3


async def async_are_contract_addresses(addresses, block='latest', provider=None):
    codes = await rpc.async_batch_eth_get_code(
        addresses=addresses, block_number=block, provider=provider,
    )
    block = list(codes.keys())[0]
    return {address: len(code) > 3 for address, code in codes[block].items()}




from __future__ import annotations

import typing

from ctc import spec
from . import rpc_registry
from . import rpc_request


def construct(
    method: str,
    *,
    batch_parameter: typing.Optional[str] = None,
    batch_values: typing.Optional[list[typing.Any]] = None,
    **parameters
) -> spec.RpcRequest:
    """create an rpc request according to a specific method"""
    constructor = rpc_registry.get_constructors()[method]
    if batch_parameter is not None and batch_values is not None:
        return [
            constructor(**{batch_parameter: value}, **parameters)
            for value in batch_values
        ]
    else:
        return constructor(**parameters)


def digest(
    response: spec.RpcRequest,
    request: spec.RpcResponse,
    digest_kwargs: typing.Optional[dict] = None,
) -> spec.RpcResponse:
    """process an rpc response"""

    if isinstance(request, dict):
        digestor = rpc_registry.get_digestors()[request['method']]
        if digest_kwargs is None:
            digest_kwargs = {}
        return digestor(response=response, **digest_kwargs)
    elif isinstance(request, list) and isinstance(response, list):
        return [
            digest(subresponse, subrequest, digest_kwargs)
            for subresponse, subrequest in zip(response, request)
        ]
    else:
        raise Exception()


def execute(
    request: spec.RpcRequest,
    provider: spec.ProviderSpec = None,
    digest_kwargs: typing.Optional[dict] = None,
) -> spec.RpcResponse:
    """send an rpc request and digest the corresponding response"""
    response = rpc_request.send(request=request, provider=provider)
    return digest(response, request=request, digest_kwargs=digest_kwargs)


async def async_execute(
    request: spec.RpcRequest,
    provider: spec.ProviderSpec = None,
    digest_kwargs: typing.Optional[dict] = None,
) -> spec.RpcResponse:
    response = await rpc_request.async_send(request=request, provider=provider)
    return digest(response, request=request, digest_kwargs=digest_kwargs)




from __future__ import annotations

import typing
from typing_extensions import TypedDict

import toolconfig

from ctc import spec
from . import config_spec


class _ToolconfigKwargs(TypedDict):
    config_path_env_var: str
    default_config_path: str


_kwargs: _ToolconfigKwargs = {
    'config_path_env_var': config_spec.config_path_env_var,
    'default_config_path': config_spec.default_config_path,
}


def get_config_path(*, raise_if_dne: bool = True) -> str:
    return toolconfig.get_config_path(raise_if_dne=raise_if_dne, **_kwargs)


def config_path_exists() -> bool:
    return toolconfig.config_path_exists(**_kwargs)


@typing.overload
def get_config(validate: typing.Literal['raise'] = 'raise') -> spec.ConfigSpec:
    ...


@typing.overload
def get_config(
    validate: typing.Literal['warn', False]
) -> typing.MutableMapping:
    ...


def get_config(
    validate: toolconfig.ValidationOption = 'raise',
) -> typing.Union[spec.ConfigSpec, typing.MutableMapping]:

    config = toolconfig.get_config(
        config_spec=spec.ConfigSpec, validate=validate, **_kwargs
    )
    if validate == 'raise':
        return typing.cast(spec.ConfigSpec, config)
    else:
        return config


def config_is_valid() -> bool:
    return toolconfig.config_is_valid(config_spec=spec.ConfigSpec, **_kwargs)




from __future__ import annotations

from ctc import binary
from ctc import spec
from . import function_parsing


def get_event_hash(event_abi: spec.EventABI) -> str:
    """compute event hash from event's abi"""
    signature = get_event_signature(event_abi=event_abi)
    return binary.keccak_text(signature)


def get_event_signature(event_abi: spec.EventABI) -> str:
    arg_types = [var['type'] for var in event_abi['inputs']]
    arg_types = [
        function_parsing.get_function_selector_type(item) for item in arg_types
    ]
    inputs = ','.join(arg_types)
    return event_abi['name'] + '(' + inputs + ')'


def get_event_unindexed_types(
    event_abi: spec.EventABI,
) -> list[spec.ABIDatumType]:
    """get list of data types in signature of event"""
    return [var['type'] for var in event_abi['inputs'] if not var['indexed']]


def get_event_unindexed_names(event_abi: spec.EventABI) -> list[str]:
    """get list of data names in signature of event"""
    return [var['name'] for var in event_abi['inputs'] if not var['indexed']]


def get_event_indexed_names(event_abi: spec.EventABI) -> list[str]:
    """get list of indexed names in signature of event"""
    return [var['name'] for var in event_abi['inputs'] if var['indexed']]


def get_event_indexed_types(
    event_abi: spec.EventABI,
) -> list[spec.ABIDatumType]:
    """get list of indexed types in signature of event"""
    return [var['type'] for var in event_abi['inputs'] if var['indexed']]




import typing


if typing.TYPE_CHECKING:
    import numpy
    import pandas

    DataFrame = pandas.core.frame.DataFrame
    Series = pandas.core.series.Series
    DType = numpy.typing.DTypeLike
    NumpyArray = numpy.typing.NDArray

else:

    DataFrame = typing.Any
    Series = typing.Any
    DType = typing.Any
    NumpyArray = typing.Any




"""

## Resources
for overview of different chains see
- https://github.com/ethereum-lists/chains
- https://chainlist.org/
"""

from __future__ import annotations

import os

from ctc import config
from ctc import spec
from ctc.toolbox import search_utils
from ctc.toolbox import store_utils


def get_network_name(network: spec.NetworkReference) -> spec.NetworkName:
    return get_network_metadata(network=network)['name']


def get_network_chain_id(network: spec.NetworkReference) -> spec.NetworkId:
    return get_network_metadata(network=network)['chain_id']


def get_network_metadata(
    network: spec.NetworkReference,
) -> spec.NetworkMetadata:

    default_network_data = load_networks_from_disk()
    config_network_data = config.get_networks()
    all_network_data = dict(default_network_data, **config_network_data)
    network_list = list(all_network_data.values())

    if network is None:
        network = config.get_default_network()

    if isinstance(network, int):
        return search_utils.get_matching_entry(
            sequence=network_list,
            query={'chain_id': network},
        )

    elif isinstance(network, str):
        return search_utils.get_matching_entry(
            sequence=network_list,
            query={'name': network},
        )

    else:
        raise Exception('unknown network format: ' + str(network))


def get_networks() -> dict[spec.NetworkName, spec.NetworkMetadata]:
    # use filesystem backend for now
    return load_networks_from_disk()


#
# # filesystem backend
#


def load_networks_from_disk(
    use_default: bool = False,
) -> dict[spec.NetworkName, spec.NetworkMetadata]:
    path = get_networks_file_path(use_default=use_default)
    raw_data = store_utils.load_file_data(path)
    return {network['name']: network for network in raw_data}


def get_networks_file_path(use_default: bool = False) -> str:
    if use_default:
        data_root = config.get_default_data_dir()
    else:
        data_root = config.get_data_dir()
    return os.path.join(data_root, 'networks.csv')




import decimal

import numpy as np

from ctc.toolbox import validate_utils


def mint_liquidity(
    x_reserves,
    y_reserves,
    lp_total_supply,
    x_deposited=None,
    y_deposited=None,
    lp_minted=None,
):

    # validate inputs
    validate_utils._ensure_exactly_one(x_deposited, y_deposited, lp_minted)

    # compute alpha
    if x_deposited is not None:
        alpha = x_deposited / x_reserves
    elif y_deposited is not None:
        alpha = y_deposited / y_reserves
    elif lp_minted is not None:
        alpha = lp_minted / lp_total_supply
    else:
        raise Exception('could not compute alpha')
    if not isinstance(alpha, decimal.Decimal) and np.isinf(alpha):
        raise Exception('initialize pool first')
    elif alpha < 0:
        raise Exception('use burn_liquidity() for negative values')

    # compute new values
    x_reserves_new = (1 + alpha) * x_reserves
    y_reserves_new = (1 + alpha) * y_reserves
    lp_total_supply_new = (1 + alpha) * lp_total_supply

    return {
        'x_deposited': x_reserves_new - x_reserves,
        'y_deposited': y_reserves_new - y_reserves,
        'lp_minted': lp_total_supply_new - lp_total_supply,
        'new_pool': {
            'x_reserves': x_reserves_new,
            'y_reserves': y_reserves_new,
            'lp_total_supply': lp_total_supply_new,
        },
    }


def burn_liquidity(
    x_reserves,
    y_reserves,
    lp_total_supply,
    x_withdrawn=None,
    y_withdrawn=None,
    lp_burned=None,
):

    # validate inputs
    validate_utils._ensure_exactly_one(x_withdrawn, y_withdrawn, lp_burned)

    # compute alpha
    if x_withdrawn is not None:
        alpha = x_withdrawn / x_reserves
    elif y_withdrawn is not None:
        alpha = y_withdrawn / y_reserves
    elif lp_burned is not None:
        alpha = lp_burned / lp_total_supply
    else:
        raise Exception('could not compute alpha')
    if alpha > 1:
        raise Exception('withdrawal exceeds pool size')
    if alpha < 0:
        raise Exception('use mint_liquidity() for negative values')

    # compute new values
    x_reserves_new = (1 - alpha) * x_reserves
    y_reserves_new = (1 - alpha) * y_reserves
    lp_total_supply_new = (1 - alpha) * lp_total_supply

    return {
        'x_withdrawn': x_reserves - x_reserves_new,
        'y_withdrawn': y_reserves - y_reserves_new,
        'lp_burned': lp_total_supply - lp_total_supply_new,
        'new_pool': {
            'x_reserves': x_reserves_new,
            'y_reserves': y_reserves_new,
            'lp_total_supply': lp_total_supply_new,
        },
    }




from __future__ import annotations

import typing

from ctc import spec
from ctc.protocols import balancer_utils
from ctc.toolbox import pd_utils

from .. import analytics_spec


async def async_compute_buybacks(
    blocks: list[int], verbose: bool = False
) -> analytics_spec.MetricGroup:

    return {
        'name': 'Buybacks',
        'metrics': {
            'buybacks_usd': (await async_compute_tribe_buybacks_usd(blocks)),
        },
    }


async def async_compute_tribe_buybacks_usd(
    blocks: list[int], swaps: typing.Optional[spec.DataFrame] = None
) -> analytics_spec.MetricData:

    # load swaps
    if swaps is None:
        swaps = await balancer_utils.async_get_pool_swaps(
            pool_address='0xc1382fe6e17bcdbc3d35f73f5317fbf261ebeecd'
        )
    swaps = typing.cast(
        spec.DataFrame,
        swaps.droplevel('transaction_index').droplevel('log_index'),
    )

    # filter tribe buys
    fei = '0x956f47f50a910163d8bf957cf5846d573e7f87ca'
    tribe_buys: typing.Any = swaps[swaps['arg__tokenOut'] == fei]  # type: ignore
    tribe_buys = tribe_buys['arg__amountOut'].map(float) / 1e18
    cummulative_tribe_buys = tribe_buys.cumsum()
    # cummulative_tribe_buys = evm.interpolate_block_series(
    #     start_block=min(blocks),
    #     pre_fill_value=0,
    #     series=cummulative_tribe_buys,
    #     end_block=max(blocks),
    # )
    cummulative_tribe_buys = pd_utils.interpolate_series(
        series=cummulative_tribe_buys,
        start_index=min(blocks),
        end_index=max(blocks),
        pre_fill_value=0,
    )

    # filter tribe sells
    tribe_sells_df = swaps[swaps['arg__tokenIn'] == fei]  # type: ignore
    if len(tribe_sells_df) > 0:
        tribe_sells = tribe_sells_df['arg__amountIn'].map(float) / 1e18
        cummulative_tribe_sells = tribe_sells.cumsum()
        # cummulative_tribe_sells = evm.interpolate_block_series(
        #     start_block=min(blocks),
        #     pre_fill_value=0,
        #     series=cummulative_tribe_sells,
        #     end_block=max(blocks),
        # )
        cummulative_tribe_sells = pd_utils.interpolate_series(
            series=cummulative_tribe_sells,
            start_index=min(blocks),
            end_index=max(blocks),
            pre_fill_value=0,
        )

        net_tribe_buys = cummulative_tribe_buys - cummulative_tribe_sells

    else:
        net_tribe_buys = cummulative_tribe_buys

    return {
        'name': 'Buybacks USD',
        'values': [net_tribe_buys[block] for block in blocks],
        'units': 'FEI',
    }




from __future__ import annotations

import types
import typing

from ctc import spec
from . import rpc_constructors
from . import rpc_digestors


def get_constructors() -> dict[str, spec.RpcConstructor]:
    constructors: dict[str, typing.Callable[..., typing.Any]] = {}
    for key in dir(rpc_constructors):
        if key.startswith('construct_'):
            candidate = getattr(rpc_constructors, key)
            if isinstance(candidate, types.FunctionType):
                constructors[key.split('construct_')[-1]] = candidate
    return constructors


def get_digestors() -> dict[str, spec.RpcDigestor]:
    digestors: dict[str, typing.Callable[..., typing.Any]] = {}
    for key in dir(rpc_digestors):
        if key.startswith('digest_'):
            candidate = getattr(rpc_digestors, key)
            if isinstance(candidate, types.FunctionType):
                digestors[key.split('digest_')[-1]] = candidate
    return digestors


def get_constructor(method: str) -> spec.RpcConstructor:
    return get_constructors()[method]


def get_digestor(method: str) -> spec.RpcDigestor:
    return get_digestors()[method]




import pandas as pd

from ctc import binary
from ctc import evm
from ctc import rpc
from ctc.cli import cli_utils


def get_command_spec():
    return {
        'f': async_calls_command,
        'args': [
            {'name': 'args', 'kwargs': {'nargs': '*'}},
            {'name': '--to-addresses', 'kwargs': {'nargs': '+'}},
            {'name': '--blocks', 'kwargs': {'nargs': '+'}},
            {'name': '--block', 'kwargs': {}},
            {'name': '--quiet', 'kwargs': {'action': 'store_true'}},
            {'name': '--output', 'kwargs': {'default': 'stdout'}},
            {'name': '--overwrite', 'kwargs': {'action': 'store_true'}},
        ],
    }


async def async_calls_command(
    args, to_addresses, blocks, block, quiet, output, overwrite
):

    if blocks is not None and to_addresses is not None:
        raise Exception('cannot specify both --blocks or --to-addresses')
    if blocks is None and to_addresses is None:
        raise Exception('must specify either --blocks or --to-addresses')

    if blocks is not None:
        if block is not None:
            raise Exception('cannot specify both --block and --blocks')

        to_address, function_name, *function_parameters = args

        block_numbers = await cli_utils.async_resolve_blocks(blocks)

        # fetch data
        results = await rpc.async_batch_eth_call(
            to_address=to_address,
            function_name=function_name,
            function_parameters=function_parameters,
            block_numbers=block_numbers,
        )

        # get output names
        function_abi = await evm.async_get_function_abi(
            contract_address=to_address,
            function_name=function_name,
        )
        output_names = binary.get_function_output_names(
            function_abi, human_readable=True
        )

        # format into dataframe
        df = pd.DataFrame(results, index=block_numbers)
        df.index.name = 'block'
        df.columns = output_names

    elif to_addresses is not None:
        if block is None:
            block = 'latest'

        function_name, *function_parameters = args

        # assert that all address functions have the same number of outputs
        function_abi = await evm.async_get_function_abi(
            contract_address=to_addresses[0],
            function_name=function_name,
        )
        n_outputs = len(function_abi['outputs'])
        for to_address in to_addresses[1:]:
            other_function_abi = await evm.async_get_function_abi(
                contract_address=to_address,
                function_name=function_name,
            )
            if len(other_function_abi['outputs']) != n_outputs:
                print('to-addresses do no have same number of function outputs')

        # fetch data
        results = await rpc.async_batch_eth_call(
            to_addresses=to_addresses,
            function_name=function_name,
            function_parameters=function_parameters,
            block_number=block,
        )

        # name based on first contract's abi
        output_names = binary.get_function_output_names(
            function_abi, human_readable=True
        )

        df = pd.DataFrame(results, index=to_addresses)
        df.index.name = 'to_address'
        df.columns = output_names

    else:
        raise Exception('must specify either --blocks or --to-addresses')

    cli_utils.output_data(data=df, output=output, overwrite=overwrite)

    await rpc.async_close_http_session()




import pytest

from ctc import binary


@pytest.mark.parametrize(
    'test',
    [
        ['0x1234', 'prefix_hex'],
        ['1234', 'raw_hex'],
        [bytes.fromhex('1234'), 'binary'],
    ],
)
def test_get_binary_format(test):
    data, binary_format = test
    assert binary.get_binary_format(data) == binary_format


@pytest.mark.parametrize(
    'test',
    [
        ['0x1234', 2],
        ['1234', 2],
        [bytes.fromhex('1234'), 2],
    ],
)
def test_get_binary_n_bytes(test):
    data, n_bytes = test
    assert binary.get_binary_n_bytes(data) == n_bytes


conversion_tests = [
    ['1234', 'raw_hex', {}, '1234'],
    ['1234', 'prefix_hex', {}, '0x1234'],
    ['1234', 'binary', {}, bytes.fromhex('1234')],
    ['0x1234', 'raw_hex', {}, '1234'],
    ['0x1234', 'prefix_hex', {}, '0x1234'],
    ['0x1234', 'binary', {}, bytes.fromhex('1234')],
    [bytes.fromhex('1234'), 'raw_hex', {}, '1234'],
    [bytes.fromhex('1234'), 'prefix_hex', {}, '0x1234'],
    [bytes.fromhex('1234'), 'binary', {}, bytes.fromhex('1234')],
]


@pytest.mark.parametrize('test', conversion_tests)
def test_convert(test):
    data, output_format, kwargs, target = test
    assert binary.convert(data, output_format, **kwargs) == target


def test_add_binary_pad():
    pass


match_tests = [
    ['1234', '5789', '1234'],
    ['1234', '0x5789', '0x1234'],
    ['1234', bytes.fromhex('5789'), bytes.fromhex('1234')],
]


@pytest.mark.parametrize('test', match_tests)
def test_match_format(test):
    format_this, like_this, target = test
    assert binary.match_format(format_this, like_this) == target




from .rpc_block_constructors import *
from .rpc_dev_constructors import *
from .rpc_log_constructors import *
from .rpc_mining_constructors import *
from .rpc_node_constructors import *
from .rpc_state_constructors import *
from .rpc_submission_constructors import *
from .rpc_transaction_constructors import *
from .rpc_whisper_constructors import *



import asyncio

from ctc import rpc

from .. import rari_abis
from . import directory_metadata
from . import token_metadata


async def async_get_pool_ctoken(comptroller, underlying, block='latest'):
    return await rpc.async_eth_call(
        to_address=comptroller,
        block_number=block,
        function_abi=rari_abis.comptroller_abis['cTokensByUnderlying'],
        function_parameters=[underlying],
    )


async def async_get_pool_ctokens(comptroller, block='latest'):
    return await rpc.async_eth_call(
        to_address=comptroller,
        block_number=block,
        function_abi=rari_abis.comptroller_abis['getAllMarkets'],
    )


async def async_get_pool_underlying_tokens(
    *, ctokens=None, comptroller=None, block='latest'
):
    if ctokens is None:
        if comptroller is None:
            raise Exception('specify comptroller')
        ctokens = await async_get_pool_ctokens(comptroller, block=block)

    coroutines = [
        token_metadata.async_get_ctoken_underlying(ctoken=ctoken)
        for ctoken in ctokens
    ]
    underlyings = await asyncio.gather(*coroutines)
    return dict(zip(ctokens, underlyings))


async def async_get_pool_oracle(comptroller, block='latest'):
    return await rpc.async_eth_call(
        to_address=comptroller,
        block_number=block,
        function_abi=rari_abis.comptroller_abis['oracle'],
    )


async def async_get_pool_name(comptroller, all_pools=None, block='latest'):
    comptroller = comptroller.lower()
    if all_pools is None:
        all_pools = await directory_metadata.async_get_all_pools(block=block)
    for pool in all_pools:
        if pool[2] == comptroller:
            return pool[0]
    else:
        raise Exception('could not find pool')




from __future__ import annotations

from typing_extensions import TypedDict
import urllib.parse

import toolcli

from ctc import rpc
from ctc import spec
from ctc import directory
from ctc.toolbox import nested_utils
from .. import config_read


class _NetworkData(TypedDict):
    networks: dict[spec.NetworkName, spec.NetworkMetadata]
    providers: dict[spec.ProviderName, spec.Provider]
    network_defaults: spec.ConfigNetworkDefaults


def setup_networks(styles: dict[str, str]) -> tuple[_NetworkData, bool]:
    print()
    print()
    toolcli.print('## Network Setup', style=styles['header'])

    print()
    print('The following networks are already installed:')
    default_networks = directory.load_networks_from_disk(use_default=True)
    for network_name, network_metadata in default_networks.items():
        print('-', network_name)

    # add new networks
    networks: dict[spec.NetworkName, spec.NetworkMetadata] = {}
    while toolcli.input_yes_or_no(
        '\nWould you like to add additional networks? ',
        style=styles['question'],
        default='no',
    ):

        go_back = False
        valid_answer = False
        while not valid_answer:
            name = toolcli.input_prompt(
                'Network name? (to go back, input a blank line) ',
                style=styles['question'],
            )
            if name == '':
                go_back = True
                break
            elif name in default_networks or name in networks:
                answer = toolcli.input_yes_or_no(
                    'Network with this name already exists. Overwrite? ',
                    style=styles['question'],
                )
                if answer:
                    valid_answer = True
            else:
                valid_answer = True
        if go_back:
            break

        network_id = toolcli.input_int(
            'Network chain_id? ', style=styles['question']
        )
        block_explorer = toolcli.input_prompt(
            'Network block explorer? ', style=styles['question']
        )

        if name in networks:
            print()
            print('Overwriting network with name', name)

        networks[name] = {
            'name': name,
            'chain_id': network_id,
            'block_explorer': block_explorer,
        }

    # set default network
    choices_set = set(networks.keys()) | set(default_networks.keys())
    choices = sorted(choices_set)
    print()
    default_network_index = toolcli.input_number_choice(
        prompt='Which network to use as default?',
        choices=choices,
        default='mainnet',
        style=styles['question'],
    )
    default_network = choices[default_network_index]

    # get providers
    all_networks = dict(default_networks, **networks)
    providers = specify_providers(
        all_networks=all_networks,
        config_networks=networks,
        default_network=default_network,
        styles=styles,
    )
    network_defaults = specify_network_defaults(
        default_network=default_network,
        providers=providers,
        styles=styles,
    )

    data: _NetworkData = {
        'networks': networks,
        'providers': providers,
        'network_defaults': network_defaults,
    }

    # check if any settings have been changed
    if config_read.config_path_exists() and config_read.config_is_valid():
        old_config = config_read.get_config()
        old_data = {key: old_config.get(key) for key in data.keys()}
        create_new_config = not nested_utils.is_equal(data, old_data)
    else:
        create_new_config = True

    return data, create_new_config


def get_default_provider_name(provider_url: str, network: str) -> str:
    hostname = urllib.parse.urlparse(provider_url).hostname
    if hostname is not None:
        hostname_pieces = hostname.split('.')
        if len(hostname_pieces) == 1:
            hostname_piece = hostname_pieces[0]
        else:
            hostname_piece = hostname_pieces[-2]
    else:
        hostname_piece = provider_url

    return hostname_piece + '_' + network


def specify_providers(
    config_networks,
    all_networks,
    default_network,
    styles,
) -> dict[spec.ProviderName, spec.Provider]:

    # load old providers
    if config_read.config_path_exists():
        try:
            old_config = config_read.get_config(validate=False)
            default_providers: dict[str, spec.Provider] = {}
            old_providers = old_config.get('providers', default_providers)
        except Exception:
            old_providers = {}
    else:
        old_providers = {}

    # check whether to keep old providers
    if len(old_providers) > 0:
        print()
        print('Currently installed providers:')
        for provider_name, provider in old_providers.items():
            network_value = provider.get('network')
            if network_value is not None:
                network_name = network_value
            else:
                network_name = 'UNKNOWN'
            print('-', provider_name, '(network=' + network_name + ')')
        print()
        if toolcli.input_yes_or_no(
            'Keep these providers? ',
            default='yes',
            style=styles['question'],
        ):
            providers = old_providers
        else:
            providers = {}
    else:
        providers = {}

    # check whether default network has a provider
    default_network_has_providers = False
    for provider_metadata in providers.values():
        if provider_metadata['network'] == default_network:
            default_network_has_providers = True
            break
    if not default_network_has_providers:
        prompt = (
            'Do you want to specify a provider node for '
            + default_network
            + '? '
        )
        print()
        if toolcli.input_yes_or_no(
            prompt=prompt, default='yes', style=styles['question']
        ):
            provider_network = default_network
            provider_url = toolcli.input_prompt(
                prompt='Provider node URL? ', style=styles['question']
            )
            default_name = get_default_provider_name(
                provider_url=provider_url, network=default_network
            )
            provider_name = toolcli.input_prompt(
                'Provider node name? ',
                style=styles['question'],
                default=default_name,
            )

            provider_spec: spec.PartialProvider = {
                'name': provider_name,
                'url': provider_url,
                'network': provider_network,
                'protocol': 'http',
                'session_kwargs': None,
                'chunk_size': None,
            }
            # should validate that this confirms to spec.ProviderSpec
            providers[provider_name] = provider_spec

    # specify additional providers
    print()
    answer = toolcli.input_yes_or_no(
        prompt='Do you want to specify any additional providers? ',
        default='no',
        style=styles['question'],
    )
    while answer:

        # specify network
        go_back = False
        valid_answer = False
        while not valid_answer:
            print()
            print('Adding new provider node...')
            provider_network = toolcli.input_prompt(
                prompt='Network of provider? (see list of networks above) ',
                style=styles['question'],
            )
            if provider_network == '':
                go_back = True
                break
            elif provider_network not in all_networks:
                print('unknown network')
            else:
                valid_answer = True
        if go_back:
            break

        # specify url
        provider_url = toolcli.input_prompt(
            prompt='Provider URL? (enter blank line to go back) ',
            style=styles['question'],
        )
        if provider_url == '':
            break
        default_name = get_default_provider_name(
            provider_url=provider_url, network=provider_network
        )

        # specify name
        valid_answer = False
        while not valid_answer:
            provider_name = toolcli.input_prompt(
                'Provider name? (enter blank line to go back) ',
                style=styles['question'],
                default=default_name,
            )
            if provider_name == '':
                go_back = True
                break
            elif provider_name in providers:
                if toolcli.input_prompt(
                    'Provider with this name already exists. Overwrite? '
                ):
                    valid_answer = True
                else:
                    continue
            else:
                valid_answer = True
        if go_back:
            break

        provider_spec = {
            'name': provider_name,
            'url': provider_url,
            'network': provider_network,
        }
        providers[provider_name] = rpc.get_provider(provider_spec)

    return providers


def specify_network_defaults(
    default_network, providers, styles
) -> spec.ConfigNetworkDefaults:

    # compile providers for each network
    providers_per_network: dict[str, list[str]] = {}
    for provider_name, provider_metadata in providers.items():
        network = provider_metadata['network']
        providers_per_network.setdefault(network, [])
        providers_per_network[network].append(provider_name)

    # get default provider for each network
    default_providers = {}
    for network in providers_per_network:
        n_providers = len(providers_per_network[network])
        if n_providers == 1:
            default_providers[network] = providers_per_network[network][0]
        elif n_providers > 1:
            answer = toolcli.input_number_choice(
                prompt='Which provider to use as default for ' + network + '?',
                choices=providers_per_network[network],
                style=styles['question'],
            )
            default_providers[network] = providers_per_network[network][answer]

    return {
        'default_network': default_network,
        'default_providers': default_providers,
    }




import os

import toolcli

from .. import config_write
from .. import config_spec


def finalize_setup(create_new_config, config, config_path, styles):

    print()
    print()
    toolcli.print('## Final Steps', style=styles['header'])

    if create_new_config:
        config_write.write_config_file(
            config_data=config,
            path=config_path,
            overwrite='prompt',
            style=styles['question'],
        )

        print()
        toolcli.print(
            'Config file created at',
            toolcli.add_style(config_path, styles['path']),
        )

    # remind user to set env var
    if config_path != os.path.expanduser(config_spec.default_config_path):
        print()
        toolcli.print(
            'Remember to set your',
            toolcli.add_style(config_spec.config_path_env_var, styles['path']),
            'environment variable to',
            toolcli.add_style(config_path, styles['path']),
        )
        print()
        toolcli.print(
            'For example, if using bash you can add the following line to your'
            + toolcli.add_style('~/.profile:', styles['path'])
        )
        toolcli.print(
            toolcli.add_style('export', styles['quote']),
            toolcli.add_style(config_spec.config_path_env_var, styles['quote'])
            + toolcli.add_style('="', styles['quote'])
            + toolcli.add_style(
                config_path.replace('"', '\\"'), styles['quote']
            )
            + toolcli.add_style('"', styles['quote']),
        )

    else:
        print()
        toolcli.print('Using default config path ' + toolcli.add_style(config_path, styles['path']))

    # print outro
    print()
    print('ctc setup complete')




import asyncio
import math

import toolstr
import tooltime
import tooltable  # type: ignore

from ctc import directory
from ctc import rpc
from ctc.protocols import chainlink_utils
from . import fuse_queries


display_names = {
    'Tetranode\'s Flavor of the Month': 'FeiRari (Fei DAO Pool)',
    'Frax & Reflexer Stable Asset Pool': 'FRAX RAI Stable Pool',
}


def sort_nested_by(nested, key, reverse=False):
    pairs = list(nested.items())
    sorted_pairs = sorted(pairs, key=lambda pair: pair[1][key], reverse=reverse)
    return dict(sorted_pairs)


def print_fuse_pool_summary(block, *, tokens_data, pool_name):

    tvl = sum(token_data['supplied'] for token_data in tokens_data.values())
    tvb = sum(token_data['borrowed'] for token_data in tokens_data.values())

    # get block data
    block_number = block['number']
    time_iso = tooltime.timestamp_to_iso(block['timestamp']).replace('T', ' ')

    headers = {
        'name': 'token',
        'supplied_tvl': 'TVL',
        'borrowed_tvl': 'TVB',
        'liquidity_tvl': 'liquidity',
        'supply_apy': 'supply %',
        'borrow_apy': 'borrow %',
        'utilization': 'util %',
    }

    tokens_data = sort_nested_by(tokens_data, 'supplied_tvl', reverse=True)

    rows = []
    for token_name, token_data in tokens_data.items():
        row = []
        for column in headers:
            datum = token_data[column]
            if column in ['utilization', 'supply_apy', 'borrow_apy']:
                datum = toolstr.format(datum, percentage=True, decimals=2)
            elif isinstance(datum, (int, float)):
                datum = toolstr.format(
                    datum,
                    prefix='$',
                    order_of_magnitude=True,
                )
            row.append(datum)
        rows.append(row)

    if pool_name in display_names:
        pool_name = display_names[pool_name]
    toolstr.print_header(pool_name)
    print('- TVL:', toolstr.format(tvl, prefix='$', order_of_magnitude=True))
    print('- TVB:', toolstr.format(tvb, prefix='$', order_of_magnitude=True))
    print('- block:', block_number)
    # print('- time:', time_iso)
    print()
    tooltable.print_table(rows, headers=headers.values())


async def print_all_pool_summary(block='latest', n_display=15):

    if block == 'latest':
        block = await rpc.async_eth_block_number()

    all_pools = await fuse_queries.async_get_all_pools(block=block)

    pools_stats = await _async_get_all_pools_stats(all_pools, block=block)

    stats_by_comptroller = {
        pool[2]: pool_stats for pool, pool_stats in zip(all_pools, pools_stats)
    }

    sorted_pools = sorted(
        all_pools[: len(stats_by_comptroller)],
        key=lambda pool: stats_by_comptroller[pool[2]]['tvl'],
        reverse=True,
    )

    total_tvl = 0
    total_tvb = 0

    rows = []
    for pool in sorted_pools[:n_display]:

        pool_stats = stats_by_comptroller[pool[2]]

        row = []

        tvb = toolstr.format(
            pool_stats['tvb'], order_of_magnitude=True, prefix='$'
        )
        tvl = toolstr.format(
            pool_stats['tvl'], order_of_magnitude=True, prefix='$'
        )

        total_tvl += pool_stats['tvl']
        total_tvb += pool_stats['tvb']

        name = pool[0]
        name = display_names.get(pool[0], pool[0])

        row.append(all_pools.index(pool))
        row.append(name)
        row.append(tvl)
        row.append(tvb)
        rows.append(row)

    block_data = await rpc.async_eth_get_block_by_number(block)
    timestamp = tooltime.timestamp_to_iso(block_data['timestamp']).replace(
        'T', ' '
    )

    toolstr.print_header('Rari Fuse')
    print(
        '- TVL:', toolstr.format(total_tvl, prefix='$', order_of_magnitude=True)
    )
    print(
        '- TVB:', toolstr.format(total_tvb, prefix='$', order_of_magnitude=True)
    )
    print('- block:', block)
    # print('- time:', timestamp)
    print()
    tooltable.print_table(rows, headers=['#', 'pool', 'TVL', 'TVB'])


async def _async_get_all_pools_stats(all_pools, block):

    n_pools = len(all_pools)
    chunk_size = 300
    n_chunks = math.ceil(n_pools / chunk_size)

    pools_stats = []
    for c in range(n_chunks):

        # print('chunk', c + 1, '/', n_chunks)

        chunk_pools = all_pools[slice(c * chunk_size, (c + 1) * chunk_size)]
        # print(slice(c * chunk_size, (c + 1) * chunk_size))

        chunk_pools_stats = [
            asyncio.create_task(
                fuse_queries.get_pool_tvl_and_tvb(
                    comptroller=pool[2], block=block
                )
            )
            for pool in chunk_pools
        ]

        chunk_pools_stats = await asyncio.gather(*chunk_pools_stats)

        pools_stats += chunk_pools_stats

    return pools_stats


async def async_get_token_multipool_stats(token, block='latest', in_usd=True):
    pools = await fuse_queries.async_get_all_pools(block=block)

    eth_price = await chainlink_utils.async_get_eth_price(block=block)
    pools_stats = [
        asyncio.create_task(
            async_get_token_pool_stats(
                token=token,
                comptroller=pool[2],
                eth_price=eth_price,
                block=block,
                in_usd=in_usd,
            )
        )
        for pool in pools
    ]
    pools_stats = await asyncio.gather(*pools_stats)
    tvl = 0
    tvb = 0
    for pool_stats in pools_stats:
        tvl += pool_stats['tvl']
        tvb += pool_stats['tvb']

    comptrollers = [pool[2] for pool in pools]
    per_pool = dict(zip(comptrollers, pools_stats))

    per_pool = sorted(
        per_pool.items(), key=lambda item: item[1]['tvl'], reverse=True
    )
    per_pool = dict(per_pool)

    pool_names = [
        asyncio.create_task(
            fuse_queries.async_get_pool_name(
                comptroller=comptroller, all_pools=pools
            )
        )
        for comptroller in per_pool.keys()
    ]

    per_pool = {k: v for k, v in per_pool.items() if v['ctoken'] is not None}
    if len(per_pool) > 0:
        blocks_per_year = None
        for p, (comptroller, pool_stats) in enumerate(per_pool.items()):

            irm = await fuse_queries.async_get_ctoken_irm(
                ctoken=pool_stats['ctoken'],
                block=block,
            )
            blocks_per_year = await fuse_queries.async_get_irm_blocks_per_year(
                irm,
                block=block,
            )

            pool_stats['supply_apy'] = await fuse_queries.async_get_supply_apy(
                ctoken=pool_stats['ctoken'],
                block=block,
                blocks_per_year=blocks_per_year,
            )
            pool_stats['borrow_apy'] = await fuse_queries.async_get_borrow_apy(
                ctoken=pool_stats['ctoken'],
                block=block,
                blocks_per_year=blocks_per_year,
            )

            pool_stats['name'] = await pool_names[p]

            # get pool index
            for pool_index, pool in enumerate(pools):
                if pool[2] == comptroller:
                    pool_stats['pool_index'] = pool_index
                    break
            else:
                raise Exception('could not detect pool index')

    return {
        'total': {'tvl': tvl, 'tvb': tvb},
        'per_pool': per_pool,
    }


async def async_get_token_pool_stats(
    token, comptroller, eth_price, block='latest', in_usd=True,
):

    ctokens = await fuse_queries.async_get_pool_ctokens(
        comptroller=comptroller, block=block
    )
    underlyings = await fuse_queries.async_get_pool_underlying_tokens(
        ctokens=ctokens, block=block
    )
    oracle = await fuse_queries.async_get_pool_oracle(
        comptroller=comptroller, block=block
    )

    tvl = 0
    tvb = 0
    matches = []
    for ctoken, underlying in underlyings.items():
        if underlying == token:
            matches.append(ctoken)
            stats = await fuse_queries.async_get_ctoken_tvl_and_tvb(
                ctoken,
                oracle,
                eth_price,
                block=block,
                in_usd=in_usd,
            )
            tvl += stats['tvl']
            tvb += stats['tvb']

    if len(matches) > 0:
        ctoken = matches[0]
    else:
        ctoken = None

    return {'tvl': tvl, 'tvb': tvb, 'ctoken': ctoken, 'matches': matches}


async def async_print_fuse_token_summary(token, block='latest', in_usd=True):

    if directory.has_erc20_metadata(symbol=token):
        token = directory.get_erc20_address(symbol=token)
    if token == '0x0000000000000000000000000000000000000000':
        symbol = 'ETH'
    else:
        symbol = directory.get_erc20_symbol(token)

    block = await rpc.async_eth_get_block_by_number(block)
    multipool_stats = await async_get_token_multipool_stats(
        token,
        block=block['number'],
        in_usd=in_usd,
    )

    if in_usd:
        prefix = '$'
    else:
        prefix = None

    include_empty = False

    headers = [
        'pool',
        'TVL',
        'TVB',
        'supply %',
        'borrow %',
        'util %',
    ]
    rows = []
    for pool_stats in multipool_stats['per_pool'].values():
        if math.isclose(pool_stats['tvl'], 0):

            if not include_empty:
                continue

            util = 0
        else:
            util = pool_stats['tvb'] / pool_stats['tvl']
        row = [
            pool_stats['name'],
            toolstr.format(
                pool_stats['tvl'], order_of_magnitude=True, prefix=prefix
            ),
            toolstr.format(
                pool_stats['tvb'], order_of_magnitude=True, prefix=prefix
            ),
            toolstr.format(pool_stats['supply_apy'], percentage=True),
            toolstr.format(pool_stats['borrow_apy'], percentage=True),
            toolstr.format(util, percentage=True),
        ]
        rows.append(row)

    tvl = 0
    tvb = 0
    for pool_stats in multipool_stats['per_pool'].values():
        tvl += pool_stats['tvl']
        tvb += pool_stats['tvb']
    timestamp = tooltime.timestamp_to_iso(block['timestamp']).replace('T', ' ')

    toolstr.print_header(symbol + ' Token Fuse Usage')
    print('- TVL:', toolstr.format(tvl, order_of_magnitude=True, prefix=prefix))
    print('- TVB:', toolstr.format(tvb, order_of_magnitude=True, prefix=prefix))
    print('- block:', block['number'])
    # print('- time:', timestamp)
    print()
    tooltable.print_table(rows, headers=headers)




from __future__ import annotations

import asyncio

import pandas as pd

from ctc import evm
from ctc import spec
from .. import analytics_spec


async def async_compute_dex_volume(
    blocks: list[int], verbose: bool = False
) -> analytics_spec.MetricGroup:

    volume_tasks = []
    for pool_name, pool_info in analytics_spec.dex_pools.items():
        platform = pool_info['platform']
        address = pool_info['address']
        if platform == 'Uniswap V2':
            task = async_compute_uniswap_v2_volume(address, blocks)
        elif platform == 'Uniswap V3':
            task = async_compute_uniswap_v3_volume(address, blocks)
        elif platform == 'Curve':
            task = async_compute_curve_volume(
                address,
                blocks,
                fei_index=pool_info['fei_index'],
                event_name=pool_info['event_name'],
            )
        elif platform == 'Sushi':
            task = async_compute_sushiswap_volume(address, blocks)
        elif platform == 'Saddle':
            fei_index = pool_info['fei_index']
            task = async_compute_saddle_volume(
                address, blocks, fei_index=fei_index
            )
        else:
            raise Exception('unknown platform: ' + str())
        volume_tasks.append(asyncio.create_task(task))

    names = list(analytics_spec.dex_pools.keys())
    volumes = await asyncio.gather(*volume_tasks)
    volume_by_pool = dict(zip(names, volumes))

    # sum by platform
    volume_by_platform = {}
    for pool_name, pool_info in analytics_spec.dex_pools.items():
        pool_volume = volume_by_pool[pool_name]
        platform = pool_info['platform']

        if platform not in volume_by_platform:
            volume_by_platform[platform] = pool_volume
            volume_by_platform[platform]['name'] = platform
        else:
            assert pool_volume['units'] == volume_by_platform[platform]['units']
            volume_by_platform[platform]['values'] = [
                sum(datapoints)
                for datapoints in zip(
                    pool_volume['values'],
                    volume_by_platform[platform]['values'],
                )
            ]

    return {
        'name': 'Volume By Platform',
        'metrics': volume_by_platform,
    }


async def async_compute_uniswap_v2_volume(
    pool_address: spec.ContractAddress, blocks: list[int]
) -> analytics_spec.MetricData:
    from ctc.protocols import uniswap_v2_utils

    swaps = await uniswap_v2_utils.async_get_pool_swaps(
        pool_address=pool_address,
        start_block=blocks[0],
        replace_symbols=True,
    )
    fei_sold = evm.bin_by_blocks(swaps['FEI_sold'], blocks)
    fei_bought = evm.bin_by_blocks(swaps['FEI_sold'], blocks)
    return {
        'values': list(fei_sold.values + fei_bought.values),
        'units': 'FEI',
    }


async def async_compute_uniswap_v3_volume(
    pool_address: spec.ContractAddress, blocks: list[int]
) -> analytics_spec.MetricData:
    import numpy as np
    from ctc.protocols import uniswap_v3_utils

    swaps = await uniswap_v3_utils.async_get_pool_swaps(
        pool_address=pool_address,
        start_block=blocks[0],
        replace_symbols=True,
    )
    binned = evm.bin_by_blocks(np.abs(swaps['FEI_amount']), blocks)
    return {
        'values': list(binned.values),
        'units': 'FEI',
    }


async def async_compute_sushiswap_volume(pool_address, blocks):
    return await async_compute_uniswap_v2_volume(pool_address, blocks)


async def async_compute_curve_volume(
    pool_address: spec.ContractAddress,
    blocks: list[int],
    fei_index: int,
    event_name: str,
) -> analytics_spec.MetricData:

    swaps = await evm.async_get_events(
        contract_address=pool_address,
        event_name=event_name,
        start_block=blocks[0],
        keep_multiindex=False,
    )

    fei_sells = swaps[swaps['arg__sold_id'] == fei_index]
    fei_sold = fei_sells['arg__tokens_sold'].map(float) / 1e18
    fei_buys = swaps[swaps['arg__bought_id'] == fei_index]
    fei_bought = fei_buys['arg__tokens_bought'].map(float) / 1e18

    # fei_bought_groups = pd.cut(fei_bought.index, blocks, right=False)
    # fei_bought_by_day = fei_bought.groupby(fei_bought_groups).sum()
    # fei_sold_groups = pd.cut(fei_sold.index, blocks, right=False)
    # fei_sold_by_day = fei_sold.groupby(fei_sold_groups).sum()
    fei_bought_by_day = evm.bin_by_blocks(data=fei_bought, blocks=blocks)
    fei_sold_by_day = evm.bin_by_blocks(data=fei_sold, blocks=blocks)
    result = fei_bought_by_day + fei_sold_by_day

    return {
        'values': list(result.values),
        'units': 'FEI',
    }


async def async_compute_saddle_volume(
    pool_address: spec.ContractAddress, blocks: list[int], fei_index: int
) -> analytics_spec.MetricData:

    swaps = await evm.async_get_events(
        contract_address=pool_address,
        event_name='TokenSwap',
        start_block=blocks[0],
        keep_multiindex=False,
    )

    fei_sells = swaps[swaps['arg__soldId'] == fei_index]
    fei_sold = fei_sells['arg__tokensSold'].map(float) / 1e18
    fei_buys = swaps[swaps['arg__boughtId'] == fei_index]
    fei_bought = fei_buys['arg__tokensBought'].map(float) / 1e18

    # fei_bought_groups = pd.cut(fei_bought.index, blocks, right=False)
    # fei_bought_by_day = fei_bought.groupby(fei_bought_groups).sum()
    # fei_sold_groups = pd.cut(fei_sold.index, blocks, right=False)
    # fei_sold_by_day = fei_sold.groupby(fei_sold_groups).sum()
    fei_bought_by_day = evm.bin_by_blocks(data=fei_bought, blocks=blocks)
    fei_sold_by_day = evm.bin_by_blocks(data=fei_sold, blocks=blocks)
    result = fei_bought_by_day + fei_sold_by_day

    return {
        'name': 'Saddle D4',
        'values': list(result.values),
        'units': 'FEI',
    }




import asyncio

from ctc import evm
from ctc import spec

from . import uniswap_v2_metadata


async def async_get_pool_swaps(
    pool_address,
    start_block=None,
    end_block=None,
    replace_symbols=False,
    normalize=True,
    provider: spec.ProviderSpec = None,
) -> spec.DataFrame:

    if normalize or replace_symbols:
        symbols_task = asyncio.create_task(
            uniswap_v2_metadata.async_get_pool_symbols(
                pool_address, provider=provider
            )
        )
        decimals_task = asyncio.create_task(
            uniswap_v2_metadata.async_get_pool_decimals(
                pool_address, provider=provider
            )
        )

    swaps = await evm.async_get_events(
        event_name='Swap',
        contract_address=pool_address,
        start_block=start_block,
        end_block=end_block,
        provider=provider,
    )

    if normalize or replace_symbols:
        symbols = await symbols_task
        decimals = await decimals_task

    # rename columns
    if replace_symbols:
        x_symbol, y_symbol = symbols
    else:
        x_symbol = 'x'
        y_symbol = 'y'
    columns = {
        'arg__amount0In': x_symbol + '_sold',
        'arg__amount0Out': x_symbol + '_bought',
        'arg__amount1In': y_symbol + '_sold',
        'arg__amount1Out': y_symbol + '_bought',
    }
    swaps = swaps.rename(columns=columns)

    # normalize columns
    if normalize:
        x_decimals, y_decimals = decimals
        swaps[columns['arg__amount0In']] = swaps[
            columns['arg__amount0In']
        ].astype(float) / (10 ** x_decimals)
        swaps[columns['arg__amount0Out']] = swaps[
            columns['arg__amount0Out']
        ].astype(float) / (10 ** x_decimals)
        swaps[columns['arg__amount1In']] = swaps[
            columns['arg__amount1In']
        ].astype(float) / (10 ** y_decimals)
        swaps[columns['arg__amount1Out']] = swaps[
            columns['arg__amount1Out']
        ].astype(float) / (10 ** y_decimals)

    return swaps


async def async_get_pool_mints(
    pool_address,
    start_block=None,
    end_block=None,
    # replace_symbols=False,
    # normalize=True,
):
    return evm.get_events(
        event_name='Mint',
        contract_address=pool_address,
        start_block=start_block,
        end_block=end_block,
    )


async def async_get_pool_burns(
    pool_address,
    start_block=None,
    end_block=None,
    # replace_symbols=False,
    # normalize=True,
):
    return evm.get_events(
        event_name='Burn',
        contract_address=pool_address,
        start_block=start_block,
        end_block=end_block,
    )




from __future__ import annotations

import os
import typing

import toolcache

import ctc.config
from ctc.toolbox import search_utils
from ctc.toolbox import store_utils
from ctc import spec


def get_oracle_address(
    name: str,
    network: spec.NetworkName = 'mainnet',
    protocol: str = 'chainlink',
    oracle_type: spec.Oracletype = 'feed',
    block: spec.BlockNumberReference = 'latest',
):

    oracle = get_oracle_metadata(
        name=name,
        network=network,
        protocol=protocol,
        oracle_type=oracle_type,
        block=block,
    )
    return oracle['address']


def get_oracle_name(
    address: spec.Address,
    network: spec.NetworkName = 'mainnet',
    protocol: str = 'chainlink',
    oracle_type: spec.Oracletype = 'feed',
    block: spec.BlockNumberReference = 'latest',
):
    oracle = get_oracle_metadata(
        address=address,
        network=network,
        protocol=protocol,
        oracle_type=oracle_type,
        block=block,
    )
    return oracle['name']


def has_oracle_metadata(
    name: typing.Optional[str] = None,
    *,
    address: typing.Optional[spec.Address] = None,
    network: spec.NetworkName = 'mainnet',
    protocol: str = 'chainlink',
) -> bool:
    try:
        get_oracle_metadata(
            name=name, address=address, network=network, protocol=protocol
        )
        return True
    except LookupError:
        return False


def get_oracle_metadata(
    name: typing.Optional[str] = None,
    *,
    address: typing.Optional[spec.Address] = None,
    network: spec.NetworkName = 'mainnet',
    protocol: str = 'chainlink',
    oracle_type: spec.Oracletype = 'feed',
    block: spec.BlockNumberReference = 'latest',
) -> spec.OracleFeedMetadata:
    if oracle_type == 'feed':
        return get_oracle_feed_metadata(
            name=name,
            address=address,
            network=network,
            protocol=protocol,
            block=block,
        )
    else:
        raise Exception('unknown oracle type: ' + str(oracle_type))


def get_oracle_feed_metadata(
    name: typing.Optional[str] = None,
    *,
    address: typing.Optional[spec.Address] = None,
    network: spec.NetworkName = 'mainnet',
    protocol: str = 'chainlink',
    block: spec.BlockNumberReference = 'latest',
) -> spec.OracleFeedMetadata:

    # block currently not used

    oracle_feeds = load_oracle_feeds(network=network, protocol=protocol)
    if name is not None:
        return search_utils.get_matching_entry(
            list(oracle_feeds.values()),
            query={'name': name},
        )
    elif address is not None:
        return search_utils.get_matching_entry(
            list(oracle_feeds.values()),
            query={'address': address.lower()},
        )
    else:
        raise Exception('must specify name or address')


#
# # backend functions
#


@toolcache.cache(cachetype='memory')
def load_oracle_feeds(
    network: spec.NetworkName = 'mainnet',
    protocol: str = 'chainlink',
) -> dict[str, spec.OracleFeedMetadata]:
    path = get_oracle_feed_path(network=network, protocol=protocol)
    oracle_feeds = store_utils.load_file_data(path)
    return {oracle_feed['name']: oracle_feed for oracle_feed in oracle_feeds}


def get_oracle_feed_path(
    network: spec.NetworkName = 'mainnet', protocol: str = 'chainlink'
) -> str:
    data_dir = ctc.config.get_data_dir()
    return os.path.join(data_dir, network, 'oracle_feeds', protocol + '.csv')

